"""Base methods."""

import importlib
import os
import sys
from distutils.version import LooseVersion

import ansible
import toolz

from ansiblelater import settings
from ansiblelater import utils


def get_settings(args):
    """
    Get new settings object.

    :param args: cli args from argparse
    :returns: Settings object

    """
    config = settings.Settings(args=args)

    return config


def get_standards(filepath):
    sys.path.append(os.path.abspath(os.path.expanduser(filepath)))
    try:
        standards = importlib.import_module("standards")
    except ImportError as e:
        utils.sysexit_with_message(
            "Could not import standards from directory {path}: {msg}".format(
                path=filepath, msg=str(e)
            )
        )

    if getattr(standards, "ansible_min_version", None) and \
            LooseVersion(standards.ansible_min_version) > LooseVersion(ansible.__version__):
        utils.sysexit_with_message(
            "Standards require ansible version {min_version} (current version {version}). "
            "Please upgrade ansible.".format(
                min_version=standards.ansible_min_version, version=ansible.__version__
            )
        )

    if getattr(standards, "ansible_later_min_version", None) and \
            LooseVersion(standards.ansible_later_min_version) > LooseVersion(
                utils.get_property("__version__")):
        utils.sysexit_with_message(
            "Standards require ansible-later version {min_version} (current version {version}). "
            "Please upgrade ansible-later.".format(
                min_version=standards.ansible_later_min_version,
                version=utils.get_property("__version__")
            )
        )

    normalized_std = (list(toolz.remove(lambda x: x.id == "", standards.standards)))
    unique_std = len(list(toolz.unique(normalized_std, key=lambda x: x.id)))
    all_std = len(normalized_std)
    if not all_std == unique_std:
        utils.sysexit_with_message(
            "Detect duplicate ID's in standards definition. Please use unique ID's only."
        )

    return standards.standards



"""Checks related to generic YAML syntax (yamllint)."""

import codecs
import os

import yaml

from ansiblelater.command.candidates import Error
from ansiblelater.command.candidates import Result
from ansiblelater.utils.rulehelper import get_action_tasks
from ansiblelater.utils.rulehelper import get_normalized_task
from ansiblelater.utils.rulehelper import get_normalized_yaml
from ansiblelater.utils.rulehelper import run_yamllint


def check_yaml_has_content(candidate, settings):
    lines, errors = get_normalized_yaml(candidate, settings)
    description = "the file appears to have no useful content"

    if not lines and not errors:
        errors.append(Error(None, description))

    return Result(candidate.path, errors)


def check_native_yaml(candidate, settings):
    tasks, errors = get_action_tasks(candidate, settings)
    description = "task arguments appear to be in key value rather than YAML format"

    if not errors:
        for task in tasks:
            normal_form, error = get_normalized_task(task, candidate, settings)
            if error:
                errors.extend(error)
                break

            action = normal_form["action"]["__ansible_module__"]
            arguments = [
                bytes(x, "utf-8").decode("unicode_escape")
                for x in normal_form["action"]["__ansible_arguments__"]
            ]
            # Cope with `set_fact` where task["set_fact"] is None
            if not task.get(action):
                continue
            if isinstance(task[action], dict):
                continue
            # strip additional newlines off task[action]
            task_action = bytes(task[action].strip(), "utf-8").decode("unicode_escape")
            if task_action.split() != arguments:
                errors.append(Error(task["__line__"], description))
    return Result(candidate.path, errors)


def check_yaml_empty_lines(candidate, settings):
    options = "rules: {{empty-lines: {conf}}}".format(conf=settings["yamllint"]["empty-lines"])
    errors = run_yamllint(candidate.path, options)
    return Result(candidate.path, errors)


def check_yaml_indent(candidate, settings):
    options = "rules: {{indentation: {conf}}}".format(conf=settings["yamllint"]["indentation"])
    errors = run_yamllint(candidate.path, options)
    return Result(candidate.path, errors)


def check_yaml_hyphens(candidate, settings):
    options = "rules: {{hyphens: {conf}}}".format(conf=settings["yamllint"]["hyphens"])
    errors = run_yamllint(candidate.path, options)
    return Result(candidate.path, errors)


def check_yaml_document_start(candidate, settings):
    options = "rules: {{document-start: {conf}}}".format(
        conf=settings["yamllint"]["document-start"]
    )
    errors = run_yamllint(candidate.path, options)
    return Result(candidate.path, errors)


def check_yaml_document_end(candidate, settings):
    options = "rules: {{document-end: {conf}}}".format(conf=settings["yamllint"]["document-end"])
    errors = run_yamllint(candidate.path, options)
    return Result(candidate.path, errors)


def check_yaml_colons(candidate, settings):
    options = "rules: {{colons: {conf}}}".format(conf=settings["yamllint"]["colons"])
    errors = run_yamllint(candidate.path, options)
    return Result(candidate.path, errors)


def check_yaml_file(candidate, settings):
    errors = []
    filename = candidate.path

    if os.path.isfile(filename) and os.path.splitext(filename)[1][1:] != "yml":
        errors.append(Error(None, "file does not have a .yml extension"))
    elif os.path.isfile(filename) and os.path.splitext(filename)[1][1:] == "yml":
        with codecs.open(filename, mode="rb", encoding="utf-8") as f:
            try:
                yaml.safe_load(f)
            except Exception as e:
                errors.append(
                    Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem))
                )

    return Result(candidate.path, errors)



"""Checks related to ansible roles files."""

from ansible.parsing.yaml.objects import AnsibleMapping
from nested_lookup import nested_lookup

from ansiblelater.command.candidates import Error
from ansiblelater.command.candidates import Result
from ansiblelater.utils.rulehelper import get_raw_yaml
from ansiblelater.utils.rulehelper import get_tasks


def check_meta_main(candidate, settings):
    content, errors = get_raw_yaml(candidate, settings)
    keys = ["author", "description", "min_ansible_version", "platforms", "dependencies"]
    description = "file should contain '{key}' key"

    if not errors:
        for key in keys:
            if not nested_lookup(key, content):
                errors.append(Error(None, description.format(key=key)))

    return Result(candidate.path, errors)


def check_scm_in_src(candidate, settings):
    roles, errors = get_tasks(candidate, settings)
    description = "usage of src: scm+url not recommended"

    if not errors:
        for role in roles:
            if isinstance(role, AnsibleMapping):
                if "+" in role.get("src"):
                    errors.append(Error(role["__line__"], description))

    return Result(candidate.path, errors)



"""Abstracted methods to simplify role writeup."""

import codecs
from collections import defaultdict

import yaml
from yamllint import linter
from yamllint.config import YamlLintConfig

from ansiblelater.command.candidates import Error
from ansiblelater.exceptions import LaterAnsibleError
from ansiblelater.exceptions import LaterError

from .yamlhelper import action_tasks
from .yamlhelper import normalize_task
from .yamlhelper import normalized_yaml
from .yamlhelper import parse_yaml_linenumbers


def get_tasks(candidate, settings):
    errors = []
    try:
        with codecs.open(candidate.path, mode="rb", encoding="utf-8") as f:
            yamllines = parse_yaml_linenumbers(f, candidate.path)

    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))
    except LaterAnsibleError as e:
        errors.append(Error(e.line, "syntax error: {msg}".format(msg=e.message)))

    return yamllines, errors


def get_action_tasks(candidate, settings):
    tasks = []
    errors = []
    try:
        with codecs.open(candidate.path, mode="rb", encoding="utf-8") as f:
            yamllines = parse_yaml_linenumbers(f, candidate.path)

        if yamllines:
            tasks = action_tasks(yamllines, candidate)
    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))
    except LaterAnsibleError as e:
        errors.append(Error(e.line, "syntax error: {}".format(e.message)))

    return tasks, errors


def get_normalized_task(task, candidate, settings):
    normalized = None
    errors = []
    try:
        normalized = normalize_task(task, candidate.path, settings["ansible"]["custom_modules"])
    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))
    except LaterAnsibleError as e:
        errors.append(Error(e.line, "syntax error: {msg}".format(msg=e.message)))

    return normalized, errors


def get_normalized_tasks(candidate, settings, full=False):
    normalized = []
    errors = []
    try:
        with codecs.open(candidate.path, mode="rb", encoding="utf-8") as f:
            yamllines = parse_yaml_linenumbers(f, candidate.path)

        if yamllines:
            tasks = action_tasks(yamllines, candidate)
            for task in tasks:
                # An empty `tags` block causes `None` to be returned if
                # the `or []` is not present - `task.get("tags", [])`
                # does not suffice.

                # Deprecated.
                if "skip_ansible_lint" in (task.get("tags") or []) and not full:
                    # No need to normalize_task if we are skipping it.
                    continue

                if "skip_ansible_later" in (task.get("tags") or []) and not full:
                    # No need to normalize_task if we are skipping it.
                    continue

                normalized.append(
                    normalize_task(task, candidate.path, settings["ansible"]["custom_modules"])
                )

    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))
    except LaterAnsibleError as e:
        errors.append(Error(e.line, "syntax error: {msg}".format(msg=e.message)))

    return normalized, errors


def get_normalized_yaml(candidate, settings, options=None):
    errors = []

    if not options:
        options = defaultdict(dict)
        options.update(remove_empty=True)
        options.update(remove_markers=True)

    try:
        yamllines = normalized_yaml(candidate.path, options)
    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))
    except LaterAnsibleError as e:
        errors.append(Error(e.line, "syntax error: {msg}".format(msg=e.message)))

    return yamllines, errors


def get_raw_yaml(candidate, settings):
    content = None
    errors = []

    try:
        with codecs.open(candidate.path, mode="rb", encoding="utf-8") as f:
            content = yaml.safe_load(f)

    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))

    return content, errors


def run_yamllint(path, options="extends: default"):
    errors = []
    try:
        with codecs.open(path, mode="rb", encoding="utf-8") as f:
            for problem in linter.run(f, YamlLintConfig(options)):
                errors.append(Error(problem.line, problem.desc))
    except LaterError as ex:
        e = ex.original
        errors.append(Error(e.problem_mark.line + 1, "syntax error: {msg}".format(msg=e.problem)))

    return errors



"""Checks related to ansible specific best practices."""

from ansiblelater.command.candidates import Error
from ansiblelater.command.candidates import Result
from ansiblelater.utils.rulehelper import get_normalized_tasks


def check_deprecated(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings, full=True)
    description = "'{old}' is deprecated and should not be used anymore. Use '{new}' instead."

    if not errors:
        for task in tasks:
            if "skip_ansible_lint" in task.get("tags"):
                errors.append(
                    Error(
                        task["__line__"],
                        description.format(old="skip_ansible_lint", new="skip_ansible_later")
                    )
                )
    return Result(candidate.path, errors)
